#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2017 jfrfonseca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Arquivo de inicializacao do pacote, exibindo e curando operacoes expostas"""


"""
# IMPORTS
"""

from py_financas._utils.unidades_brasileiras import ultimo_dia_util, e_dia_util
from .autenticacao import inicializa_cliente_wsdl_cvm
from .arquivos_zip import le_arquivo_zip_de_string
from py_financas._utils.autenticacao import LoggingWebServicePlugin
from .unidades_brasileiras import normaliza_cnpj, normaliza_numerico





# STDLib
from typing import List, Dict, Tuple

# PIP
from pandas import read_excel, to_datetime

# This package
from py_financas.sinacor.parser import parse_notas_corretagem
from py_financas.sinacor.types import Operacao, Posicao, NotaCorretagem


def parse_operacoes_manuais(arquivo_excel_operacoes_manuais:str) -> Dict[Tuple[str,str],Operacao]:

    # Read and normalize the data
    df_operacoes_manuais = read_excel(arquivo_excel_operacoes_manuais)
    df_operacoes_manuais['data'] = to_datetime(df_operacoes_manuais['data']).dt.date
    df_operacoes_manuais = df_operacoes_manuais.sort_values('data').reset_index(drop=True).copy(deep=True)
    df_operacoes_manuais['quantidade'] = df_operacoes_manuais['quantidade'].astype(float)
    df_operacoes_manuais['valor_unitario'] = df_operacoes_manuais['valor_unitario'].astype(float)
    df_operacoes_manuais['valor_total'] = df_operacoes_manuais['valor_unitario'] * df_operacoes_manuais['quantidade']
    df_operacoes_manuais['cnpj_corretora'] = df_operacoes_manuais['cnpj_corretora'].astype(str).str.split().str[0].astype(str)

    # Parse each operation
    operacoes_manuais = {}
    for idx, linha in df_operacoes_manuais.dropna().iterrows():
        data = linha['data']
        cnpj = str(linha['cnpj_corretora']).split('.')[0]
        op = Operacao(
            _nota_numero=int(idx),
            _nota_data_pregao=data,
            _nota_corretora_cnpj=cnpj,

            titulo=linha['titulo'],
            preco_ajuste=linha['valor_unitario'],
            quantidade=linha['quantidade'],
            valor_operacao_ajuste=linha['valor_total'],
            compra_venda=linha['compra_ou_venda'],
            debito_credito='debito' if (linha['compra_ou_venda'] == 'compra') else 'credito',
            tipo_mercado=linha['tipo_mercado'],
            negociacao='bovespa',
            especificacao_titulo=('','',''),
            observacao=linha['observacoes']
        )
        key = (cnpj, op.chave_titulo)
        if key not in operacoes_manuais: operacoes_manuais[key] = []
        operacoes_manuais[key].append(op)

    return operacoes_manuais


def parse_posicoes_de_operacoes(todas_as_operacoes:Dict[Tuple[str,str],List[Operacao]]) -> List[Posicao]:

    # Parse the positions
    posicoes: Dict[Tuple[str,str],List[Posicao]] = {}
    for chave, operacoes in todas_as_operacoes.items():
        cnpj, titulo = chave
        cnpj = str(cnpj).split('.')[0]
        operacoes = sorted(operacoes, key=lambda op: op._nota_data_pregao)

        # Criando a posição inicial
        if chave not in posicoes:
            op = operacoes[0]
            posicoes[chave] = [
                Posicao(
                    corretora=str(cnpj).split('.')[0],
                    chave_operacao_abertura=op.chave,
                    titulo=op.titulo,
                    especificacao_titulo=op.especificacao_titulo,
                    observacao_titulo=op.observacao,
                    tipo_mercado=op.tipo_mercado,

                    data_abertura=op._nota_data_pregao,
                    data_fechamento=None,

                    quantidade=op.quantidade,
                    valor_unitario=op.preco_ajuste
                )
            ]
            operacoes = operacoes[1:]  # Evitamos inserir a primeira OP novamente

        # Atualizando a posição a partir das operações
        for op in operacoes:
            if posicoes[chave][-1].aberta:
                posicoes[chave][-1].atualizar(op)
                continue

            # Criacao de nova posicao do titulo, após fechamento
            else:
                posicoes[chave].append(
                    Posicao(
                        corretora=cnpj,
                        chave_operacao_abertura=op.chave,
                        titulo=op.titulo,
                        especificacao_titulo=op.especificacao_titulo,
                        observacao_titulo=op.observacao,
                        tipo_mercado=op.tipo_mercado,

                        data_abertura=op._nota_data_pregao,
                        data_fechamento=None,

                        quantidade=op.quantidade,
                        valor_unitario=op.preco_ajuste
                    )
                )

    return sorted(sum(posicoes.values(), []))


def parse_posicoes_de_notas_de_corretagem(obj_notas_corretagem:List[NotaCorretagem]) -> List[Posicao]:

    # Extract the operations from the automatic files
    operacoes_automaticas = {}
    for nota in sorted(obj_notas_corretagem):
        for folha in nota.operacoes:
            for op in folha:
                chave = (str(nota.corretora.cnpj).split('.')[0],op.chave_titulo)
                if chave in operacoes_automaticas:
                    operacoes_automaticas[chave].append(op)
                else: operacoes_automaticas[chave] = [op]

    return parse_posicoes_de_operacoes(operacoes_automaticas)


def parse_posicoes(arquivos_pdf_notas_corretagem:List[str],
                   arquivos_excel_operacoes_manuais:List[str]) -> List[Posicao]:

    # Parse all the Notas de Corretagem as NotaCorretagem objects
    obj_notas_corretagem = parse_notas_corretagem(arquivos_pdf_notas_corretagem)

    # Extract the operations from the automatic files
    operacoes_automaticas = {}
    for nota in sorted(obj_notas_corretagem):
        for folha in nota.operacoes:
            for op in folha:
                chave = (str(nota.corretora.cnpj).split('.')[0],op.chave_titulo)
                if chave in operacoes_automaticas:
                    operacoes_automaticas[chave].append(op)
                else: operacoes_automaticas[chave] = [op]

    # Extract the operations from the manual files
    operacoes_manuais = {}
    for arq in arquivos_excel_operacoes_manuais:
        for k,v in parse_operacoes_manuais(arq).items():
            if k in operacoes_manuais:
                operacoes_manuais[k].extend(v)
            else: operacoes_manuais[k] = v

    # Merge all operations
    todas_as_operacoes = {}
    todas_as_operacoes.update(operacoes_manuais)
    for k,v in operacoes_automaticas.items():
        todas_as_operacoes[k] = todas_as_operacoes.get(k, []) + v
    todas_as_operacoes = {k:sorted(v, key=lambda op: op._nota_data_pregao)
                          for k,v in todas_as_operacoes.items()}

    return parse_posicoes_de_operacoes(todas_as_operacoes)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2017 jfrfonseca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Operacoes de recuperacao de cadastros de fundos da CVM"""


"""
# IMPORTS
"""


# Biblioteca padrao
import logging

# Dependencias PIP
import requests
import xmltodict
import numpy as np
import pandas as pd

# Dependencias locais
from py_financas import _utils


"""
# PAYLOAD
"""


def recupera_cadastros(usuario_cvm, senha_cvm,
                       data_busca=_utils.unidades_brasileiras.ultimo_dia_util(),
                       justificativa='Obtencao de cadastros de cvm'):

    # Inicializamos o objeto de log
    log = logging.getLogger('py_financas:cvm')

    # Inicializamos um cliente com a CVM
    cliente_wsdl_cvm = _utils.inicializa_cliente_wsdl_cvm(usuario_cvm, senha_cvm)

    # Solicitamos uma URL para download do documento de cadastros dos cvm
    url_documento = cliente_wsdl_cvm.service.solicAutorizDownloadCadastro(data_busca.strftime("%Y-%m-%d"), justificativa)

    log.debug("Obtida a URL de download de cadastros de cvm ({}) para o dia {}".format(url_documento, data_busca))

    # Recuperamos o documento da URL
    documento_fundos_raw = requests.get(url_documento).content

    log.debug("Recuperado o documento de cadastros de cvm de {}".format(data_busca))

    # Extraimos o conteudo do documento (Zipado) para uma string XML
    documento_fundos = _utils.le_arquivo_zip_de_string(documento_fundos_raw)

    # Convertemos o documento XML para um dicionario Python
    dict_fundos = xmltodict.parse(documento_fundos)

    # Convertemos o dicionario para um objeto PANDAS
    cadastros_fundos = pd.DataFrame().from_dict(dict_fundos['ROOT']['PARTICIPANTES']['CADASTRO'])

    log.debug("Obtidos os cadastros de {} cvm".format(len(cadastros_fundos.index)))

    # Normalizamos os valores booleanos dos cadastros
    cadastros_fundos = cadastros_fundos.replace(u'Sim', True)
    cadastros_fundos = cadastros_fundos.replace(u'Não', False)

    # Inicializamos buffers para a taxa de performance e seus detalhes
    lista_taxa_performance = []
    lista_detalhes_taxa_performance = []

    # Para cada elemento da coluna da taxa de performance
    for tp in cadastros_fundos['TAXA_PERFORMANCE'].fillna('0.0'):

        # Dividimos o elemento e capturamos sua primeira parte (numerica)
        divisao = tp.split(' ', 1)
        tp_numerica = divisao[0]

        # Se houver, capturamos a segunda parte (textual)
        if len(divisao) > 1:
            detalhe = divisao[1]
        else:
            # Se nao houver, a parte textual e um np.NaN
            detalhe = np.NaN

        # Inserimos as duas partes normalizadas nos buffers
        lista_taxa_performance.append(float(tp_numerica.replace(',', '.')))
        lista_detalhes_taxa_performance.append(detalhe)

    # Aplicamos normalizações no dataframe por coluna
    cadastros_fundos = cadastros_fundos.assign(

        # Parseamos as colunas de data como objetos DATETIME (naive)
        DT_INICIO = pd.to_datetime(cadastros_fundos['DT_INICIO']),
        DT_CONSTITUICAO = pd.to_datetime(cadastros_fundos['DT_CONSTITUICAO']),
        DT_INICIO_CLASSE = pd.to_datetime(cadastros_fundos['DT_INICIO_CLASSE']),

        # Normalizamos os campos CNPJ como strings de 14 digitos numericos, inserindo zeros a esquerda
        CNPJ = cadastros_fundos['CNPJ'].apply(_utils.normaliza_cnpj),
        CNPJ_ADMINISTRADOR = cadastros_fundos['CNPJ_ADMINISTRADOR'].apply(_utils.normaliza_cnpj),

        # Normalizamos a taxa de performance como valores numericos, inserindo taxa 0 nos valores numericos
        TAXA_PERFORMANCE = lista_taxa_performance,

        # Normalizamos o texto detahado da taxa de performance, inserindo strings vazias nas posicoes faltantes
        DETALHE_TAXA_PERFORMANCE = lista_detalhes_taxa_performance
    )

    # Retornamos o dataframe, com sorte sem problemas de normalização
    return cadastros_fundos



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2017 jfrfonseca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Operacoes de recuperacao de dados diarios de fundos"""


"""
# IMPORTS
"""


# Biblioteca padrao
import json
import logging

# Dependencias PIP
import xmltodict
import requests
import pandas as pd

# Este pacote
from py_financas import _utils, constantes


"""
# PAYLOAD
"""


def recupera_informes_diarios_de_hoje(usuario_cvm, senha_cvm,
                                      justificativa='Obtencao de informes diarios de cvm'):

    # Inicializamos o objeto de log
    log = logging.getLogger('py_financas:cvm')

    # Testamos se o dia de hoje e um dia util
    if not _utils.unidades_brasileiras.e_dia_util():

        # Se nao, interrompemos a funcao imediatamente
        log.debug("Hoje nao e um dia util. Nao e possivel recuperar os informes diarios de cvm de hoje do sistema CVM")
        return pd.DataFrame()

    # Inicializamos um cliente com a CVM
    cliente_wsdl_cvm = _utils.inicializa_cliente_wsdl_cvm(usuario_cvm, senha_cvm)

    # Solicitamos uma URL para download do documento de informes diarios dos cvm
    url_documento = cliente_wsdl_cvm.service.solicAutorizDownloadArqEntrega(constantes.codigo_informes_diarios_fundos,
                                                                            justificativa)

    log.debug("Obtida a URL para download dos informes diarios de cvm: {}".format(url_documento))

    # Recuperamos o documento da URL
    documento_fundos_raw = requests.get(url_documento).content

    log.debug("Recuperado o documento de informes diarios de cvm do ultimo dia util")

    # Extraimos o conteudo do documento (Zipado) para uma string XML
    documento_fundos = _utils.le_arquivo_zip_de_string(documento_fundos_raw)

    # Convertemos o documento XML para um dicionario Python
    dict_fundos = xmltodict.parse(documento_fundos)

    # Obtemos o valor dos informes. Se um campo interno for nulo, retornamos um dataframe vazio
    valor = dict_fundos['ROOT']
    if valor is not None:
        valor = valor['INFORMES']
        if valor is not None:
            valor = valor['INFORME_DIARIO']
        else:
            return pd.DataFrame()
    else:
        return pd.DataFrame()

    # Se chegamos ate aqui, devemos continuar.
    # Normalizamos o tamanho minimo da lista
    if not isinstance(valor, list):
        valor = [valor]

    # Convertemos o dicionario para um objeto PANDAS
    dataframe_informes_fundos = pd.DataFrame().from_dict(valor)

    # Normalizamos varias colunas no dataframe
    dataframe_informes_fundos = dataframe_informes_fundos.assign(

        # Normalizamos as colunas numericas
        CAPTC_DIA = pd.to_numeric(dataframe_informes_fundos['CAPTC_DIA'].apply(_utils.normaliza_numerico)),
        NR_COTST = pd.to_numeric(dataframe_informes_fundos['NR_COTST'].apply(_utils.normaliza_numerico)),
        PATRIM_LIQ = pd.to_numeric(dataframe_informes_fundos['PATRIM_LIQ'].apply(_utils.normaliza_numerico)),
        RESG_DIA = pd.to_numeric(dataframe_informes_fundos['RESG_DIA'].apply(_utils.normaliza_numerico)),
        VL_QUOTA = pd.to_numeric(dataframe_informes_fundos['VL_QUOTA'].apply(_utils.normaliza_numerico)),
        VL_TOTAL = pd.to_numeric(dataframe_informes_fundos['VL_TOTAL'].apply(_utils.normaliza_numerico)),

        # Normalizamos as colunas de datas
        DT_COMPTC = pd.to_datetime(dataframe_informes_fundos['DT_COMPTC']),

        # Normalizamos a coluna de CNPJ
        CNPJ_FDO = dataframe_informes_fundos['CNPJ_FDO'].append(_utils.normaliza_cnpj)
    )

    # Retornamos o dataframe normalizado
    return dataframe_informes_fundos



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2017 jfrfonseca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Operacoes de autenticacao utilizando sistemas WSDL com o pacote SUDS-JURKO"""


"""
# IMPORTS
"""


# Biblioteca padrão
import logging

# Dependencias PIP
from suds.client import Client
from suds.plugin import MessagePlugin

# Este pacote
from py_financas import constantes


"""
# UTIL
"""


class LoggingWebServicePlugin(MessagePlugin):
    """
    Classe de compatibilidade SUDS-JURKO
    Fonte: https://bitbucket.org/jurko/suds/issues/39/missing-last_sent-and-last_received
    """

    def __init__(self):
        self.last_sent_message = None
        self.last_received_reply = None

    def sending(self, context):
        if context.envelope:
            self.last_sent_message = context.envelope

    def parsed(self, context):
        if context.reply:
            self.last_received_reply = context.reply

    def last_sent(self):
        return self.last_sent_message

    def last_received(self):
        return self.last_received_reply


def inicializa_cliente_wsdl_cvm(usuario_cvm, senha_cvm):

    log = logging.getLogger('py_financas:cvm')

    try:

        # Inicializamos um objeto de compatibilidade SUDS
        logging_plugin = LoggingWebServicePlugin()

        # Iniciamos um objeto cliente SOAP SUDS-JURKO
        cliente = Client(constantes.wsdl_cvm, plugins=[logging_plugin])

        # Executamos o metodo de login da CVM
        cliente.service.Login(usuario_cvm, senha_cvm)

        # Setamos os headers de login na CVM
        cliente.set_options(soapheaders=logging_plugin.last_received().getChildren()[0][0][0])

        # Retornamos o cliente formado
        return cliente

    except Exception as exp:
        log.error("Excecao no login do sistema WSDL CVM! {}".format(exp))
        raise exp



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2017 jfrfonseca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Operacoes de conversao de unidades e notacoes brasileiras em notacao cientifica/estado-unidense"""


"""
# IMPORT
"""


import datetime


"""
# UTIL
"""


# FORMATACAO DE DADOS ==================================================================================================


def normaliza_cnpj(string_nao_normalizada):
    return string_nao_normalizada.replace(".", "").replace("-", "").replace("/", "").zfill(14)


def normaliza_numerico(string_nao_normalizada):
    return string_nao_normalizada.replace(",", ".")


# DATAS ================================================================================================================


def ultimo_dia_util(data_inicial=datetime.datetime.today()):

    # Verificamos se a data passada e um dia util. Se nao, subtraimos um dia ate que seja
    while not e_dia_util(data_inicial):
        data_inicial -= datetime.timedelta(days=1)

    return data_inicial


def e_dia_util(data=datetime.datetime.today()):
    return data.weekday() < 5



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2017 jfrfonseca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Inicializacao do pacote, oferecendo multiplos alias para os sub-pacotes"""


"""
# IMPORTS
"""


__version__ = '1.0.3'
version = __version__
VERSION = __version__


# O sistema da Comissao de Valores Mobiliarios e tambem a unica fonte de informacoes sobre fundos
from py_financas import cvm
from py_financas import cvm as fundos
from py_financas import cvm as fundos_mutuos

# O sistema do Banco Central do Brasil - Sistema de Gestao de Series Temporais e tambem a unica fonte de indices
from py_financas import bcb_sgs
from py_financas import bcb_sgs as indices
from py_financas import bcb_sgs as indexadores  # Indices tambem sao chamados "indexadores"

# O SINACOR contém os modelos de dados utilizados em negociacoes
from py_financas import sinacor
from py_financas.sinacor.types import NotaCorretagem, Posicao, \
    MomentoPosicao, Operacao, Corretora, Totais, Impostos, Custos




# STDLib
from time import time
from itertools import cycle
from datetime import date, timedelta
from typing import Literal, Tuple, List, Dict, Union

# PIP
from pydantic.dataclasses import dataclass, Field


# CONSTANTS
#######################################################################################################################


DC = Literal['debito', 'credito']


# UTILS
#######################################################################################################################


def validate_cnpj(cnpj: str) -> bool:
    if not cnpj.isdigit(): raise TypeError('ONLY CNPJ NUMERALS ARE ALLOWED')
    if len(cnpj) != 14: return False  # All CNPJ codes must have 14 numerals
    if len(set(cnpj)) == 1: return False  # 14 repetitions not allowed

    cnpj_r = cnpj[::-1]
    for i in range(2, 0, -1):
        cnpj_enum = zip(cycle(range(2, 10)), cnpj_r[i:])
        dv = sum(map(lambda x: int(x[1]) * x[0], cnpj_enum)) * 10 % 11
        if cnpj_r[i - 1:i] != str(dv % 10):
            return False

    return True


def validate_positive(key:str, value:float):
    assert value >= 0, f'{key} VALUE {value} SHOULD NOT BE NEGATIVE'


def validate_equal(key_a:str, value_a:float, key_b:str, value_b:float):
    assert value_a == value_b, f'{key_a} VALUE {value_a} SHOULD EQUAL {key_b} VALUE {value_b}'


def to_float(val:str, empty_string_equals=0.0) -> float:
    if val == '': return empty_string_equals
    return float(val.strip().replace('.', '').replace(',', '.'))


def to_DC(val:str, default='') -> str: return {'D': 'debito', 'C': 'credito'}.get(val.strip().upper(), default)


def printreal(_valor:float) -> str:
    valor = str(round(_valor,2))
    if '.' in valor:
        reais, centavos = valor.split('.')
        while len(centavos) < 2: centavos += '0'
    else:
        reais = valor
        centavos = '00'
    result = ','+centavos
    while len(reais) > 0:
        result = '.'+reais[-3:]+result
        reais = reais[:-3]
    return result[1:]


# DATA CLASSES (TYPES)
#######################################################################################################################


@dataclass
class Corretora():
    nome: str
    cnpj: str
    estado: Literal['BA', 'DF', 'MG', 'PE', 'RJ', 'RS', 'SP']

    def __repr__(self) -> str: return self.nome
    def __str__(self) -> str: return self.__repr__()

    def __post_init__(self):
        assert validate_cnpj(self.cnpj), f'INVALID CNPJ: {self.cnpj}'


@dataclass
class Operacao():
    _nota_numero: int
    _nota_data_pregao: date
    _nota_corretora_cnpj: str

    titulo: str
    preco_ajuste: float
    quantidade: float
    valor_operacao_ajuste: float
    compra_venda: Literal['compra', 'venda', 'desdobramento']
    debito_credito: DC
    tipo_mercado: Literal['vista', 'fracionario', 'prazo']
    negociacao: Literal['bovespa', 'bovespa 1']
    especificacao_titulo: Tuple[str,str,str]
    observacao: str

    @property
    def chave_titulo(self) -> str:
        tm = self.tipo_mercado
        if tm == 'fracionario': tm = 'vista'
        return f'{self.titulo}:{tm}'
        #chave = ' '.join([
        #    esp.strip()
        #    for esp in list(op.especificacao_titulo) + [self.observacao]
        #    if len(esp.strip()) > 0
        #])
        #return f'{self.titulo}:{chave}:{self.tipo_mercado}'

    @property
    def chave(self) -> str:
        return f'{self._nota_data_pregao}:{self._nota_corretora_cnpj}:{self._nota_numero}:{self.chave_titulo}:{self.compra_venda}:{self.quantidade}:{self.preco_ajuste}'


@dataclass
class Totais():
    liquido: float
    liquido_DC: DC
    vendas_vista: float
    compras_vista: float
    opcoes_compras: float
    opcoes_vendas: float
    debentures: float
    operacoes: float
    operacoes_termo_bolsa: float
    operacoes_titulos_publicos: float
    valor_liquido_operacoes: float
    valor_liquido_operacoes_DC: DC
    cblc: float
    cblc_DC: DC
    bovespa_soma: float
    bovespa_soma_DC: DC
    corretagem_despesas: float
    corretagem_despesas_DC: DC

    def __post_init__(self):
        validate_positive('COMPRAS VISTA', self.compras_vista)
        validate_positive('VENDAS VISTA', self.vendas_vista)
        validate_positive('OPCOES COMPRAS', self.opcoes_compras)
        validate_positive('OPCOES VENDAS', self.opcoes_vendas)
        validate_positive('DEBENTURES', self.debentures)
        validate_positive('OPERACOES', self.operacoes)
        validate_positive('OPERACOES TERMO BOLSA', self.operacoes_termo_bolsa)
        validate_positive('OPERACOES TITULOS PUBLICOS', self.operacoes_titulos_publicos)


@dataclass
class Custos():
    taxa_liquidacao: float
    taxa_liquidacao_DC: DC
    taxa_registro: float
    taxa_registro_DC: DC
    taxa_termo_opcoes: float
    taxa_termo_opcoes_DC: DC
    taxa_ana: float
    taxa_ana_DC: DC
    emolumentos: float
    emolumentos_DC: DC
    clearing: float
    clearing_DC: DC
    execucao: float
    execucao_DC: DC
    execucao_casa: float
    execucao_casa_DC: DC
    outras: float
    outras_DC: DC

    @staticmethod
    def null() -> 'Custos': return Custos(0,'debito',0,'debito',0,'debito',
                                          0,'debito',0,'debito',0,'debito',
                                          0,'debito',0,'debito',0,'debito')


@dataclass
class Impostos():
    iss: float
    iss_DC: DC
    irrf: float
    irrf_DC: DC
    irrf_base: float

    def __post_init__(self):
        validate_positive('IRRF BASE', self.irrf_base)

    @staticmethod
    def null() -> 'Impostos': return Impostos(0,'debito',0,'debito',0)


@dataclass
class NotaCorretagem():
    _raw_content: str

    numero : int
    numero_nota_substitutiva: int
    data_pregao: date
    data_liquidacao: date
    corretora: Corretora
    cliente: str
    quantidade_folhas: int

    operacoes: List[List[Operacao]]

    totais: Totais
    custos: Custos
    impostos: Impostos

    def __post_init__(self):
        assert self.data_pregao <= self.data_liquidacao, f'DATA DO PREGAO {self.data_pregao} NAO PODE SER MAIOR QUE DATA DA LIQUIDACAO {self.data_liquidacao}'
        assert (self.data_liquidacao - self.data_pregao) <= timedelta(days=7), f'DATA DO PREGAO {self.data_pregao} MUITO ANTERIOR A DATA DA LIQUIDACAO {self.data_liquidacao}'
        assert self.quantidade_folhas == len(self.operacoes), f'QUANTIDADE DE FOLHAS INCONSISTENTE COM A QUANTIDADE DE CONJUNTOS DE OPERACOES'

    def __len__(self) -> int:
        return sum([len(pg) for pg in self.operacoes])

    @property
    def key(self) -> str:
        return self.data_pregao.strftime(f'%Y-%m-%d {self.corretora} {self.numero}')

    def __repr__(self) -> str:
        return f'<NC {self.key} [{len(self)}] ${self.totais.liquido} {self.totais.liquido_DC}>'

    def __str__(self) -> str:
        return self.__repr__()

    def __eq__(self, other) -> bool: return self.key == other.key
    def __lt__(self, other) -> bool: return self.key < other.key

    @staticmethod
    def manual(
        data_pregao:date, corretora:Corretora, cliente:str,
        operacoes:List[Operacao], totais:Totais,
        impostos:Impostos=None, custos:Custos=None,  #type:ignore
        data_liquidacao=None,
        numero=None, numero_nota_substitutiva=None,
        quantidade_folhas=1,
        _raw_content='Nota de corretagem inserida manualmente',
    ) -> 'NotaCorretagem':
        if numero is None: numero = int(time())
        if data_liquidacao is None: data_liquidacao = data_pregao
        if numero_nota_substitutiva is None: numero_nota_substitutiva = numero
        if impostos is None: impostos = Impostos.null()
        if custos is None: custos = Custos.null()
        if totais is None: totais = Totais.null()  #type:ignore
        return NotaCorretagem(
            _raw_content=_raw_content, numero=numero,
            data_pregao=data_pregao, data_liquidacao=data_liquidacao,
            numero_nota_substitutiva=numero_nota_substitutiva,
            corretora=corretora, cliente=cliente,
            quantidade_folhas=quantidade_folhas,
            operacoes=operacoes, impostos=impostos, custos=custos, totais=totais  #type:ignore
        )


@dataclass
class MomentoPosicao():
    chave_operacao: str
    data: date
    quantidade: float
    valor_unitario: float
    rendimento: float

    @property
    def valor_total(self) -> float: return round(self.quantidade * self.valor_unitario, 2)


@dataclass
class Posicao():
    corretora: str
    titulo: str
    especificacao_titulo: Tuple[str, str, str]
    observacao_titulo: str
    tipo_mercado: str
    chave_operacao_abertura: str

    data_abertura: date
    data_fechamento: Union[date, None]

    quantidade: float
    valor_unitario: float

    historico: List[MomentoPosicao] = Field(default_factory=list)
    fechamentos: Dict[date,float] = Field(default_factory=dict)

    def __post_init__(self):
        self.historico.append(
            MomentoPosicao(
                self.chave_operacao_abertura,
                self.data_abertura,
                self.quantidade,
                self.valor_unitario,
                0.0
            )
        )

    @property
    def aberta(self) -> bool: return self.data_fechamento is None

    @property
    def valor_total(self) -> float: return round(self.valor_unitario * self.quantidade, 2)

    @property
    def rendimento(self) -> float: return round(self.historico[-1].rendimento, 2)

    @property
    def chave_titulo(self) -> str:
        return f'{self.titulo}:{self.tipo_mercado}'
        #chave = ' '.join([
        #    esp.strip()
        #    for esp in list(op.especificacao_titulo) + [self.observacao_titulo]
        #    if len(esp.strip()) > 0
        #])
        #return f'{self.titulo}:{chave}:{self.tipo_mercado}'

    @property
    def chave(self) -> str:
        return f'{self.corretora}:{self.chave_titulo}:{self.data_abertura}:{self.data_fechamento}'

    def __repr__(self) -> str:
        return f'<POS {self.chave} [{len(self.historico)}|{len(self.fechamentos)}] ${self.valor_total}={round(self.quantidade, 2)}x{round(self.valor_unitario, 2)} ({self.rendimento})>'

    def __str__(self) -> str: return self.__repr__()

    def __eq__(self, other) -> bool: return self.__repr__() == other.__repr__()
    def __lt__(self, other) -> bool:
        for attr in ['data_abertura',
                     'corretora', 'chave_titulo',
                     'data_fechamento',
                     'valor_total', 'rendimento']:
            ours = getattr(self, attr)
            theirs = getattr(other, attr)
            if ours != theirs:
                return ours < theirs
        return self.__repr__ < other.__repr__

    def atualizar(self, op:Operacao) -> bool:
        data_pregao = op._nota_data_pregao

        # Sanity checks
        assert op._nota_corretora_cnpj == self.corretora, f'TENTATIVA DE ATUALIZAR UMA POSIÇÃO DE OUTRA CORRETORA! {self.chave}'
        assert op.chave_titulo == self.chave_titulo, f'TENTATIVA DE ATUALIZAR POSIÇÃO COM OPERAÇÃO DE OUTRO TÍTULO! {self.chave}'
        assert self.aberta, f'TENTATIVA DE ATUALIZAR UMA POSICÃO FECHADA! {self.chave}'
        assert data_pregao >= self.data_abertura, f'TENTATIVA DE ATUALIZAR UMA POSIÇÃO COM UMA OPERAÇÃO MAIS ANTIGA! {self.chave}'

        # Se estamos numa COMPRA, atualizamos a quantidade e o valor unitário médio
        if op.compra_venda == 'compra':
            valor_agregado = abs(op.valor_operacao_ajuste)
            quantidade_agregada = abs(op.quantidade)

            valor_atual = self.valor_total
            quantidade_atual = self.quantidade

            novo_valor = valor_atual + valor_agregado
            nova_quantidade = self.quantidade + quantidade_agregada
            novo_valor_unitario = novo_valor / nova_quantidade

            rendimento_agregado = 0.0
            self.valor_unitario = novo_valor_unitario
            self.quantidade = nova_quantidade

        # Se estamos numa venda:
        #   - Atualizamos apenas a quantidade
        #   - Criamos um novo rendimento / fechamento parcial
        #   - Se a quantidade vai a 0, fechamos a posição
        elif op.compra_venda == 'venda':
            valor_unitario_recebido = abs(op.preco_ajuste)
            quantidade_removida = abs(op.quantidade)

            valor_unitario_atual = self.valor_unitario
            quantidade_atual = self.quantidade

            rendimento_agregado = (valor_unitario_recebido - valor_unitario_atual) * quantidade_removida
            nova_quantidade = self.quantidade - quantidade_removida

            self.quantidade = nova_quantidade
            self.fechamentos[data_pregao] = self.fechamentos.get(data_pregao, 0.0) + rendimento_agregado

            if self.quantidade == 0:
                self.data_fechamento = data_pregao

        # Se estamos num DESDOBRAMENTO, atualizamos a quantidade e o valor unitário médio
        elif op.compra_venda == 'desdobramento':
            quantidade_agregada = (
                int(self.quantidade * op.preco_ajuste)
                - self.quantidade
            )

            novo_valor = self.valor_total + 0.0
            nova_quantidade = self.quantidade + quantidade_agregada
            novo_valor_unitario = novo_valor / nova_quantidade

            rendimento_agregado = 0.0
            self.valor_unitario = novo_valor_unitario
            self.quantidade = nova_quantidade

        else: raise KeyError(f'Tipo desconhecido de operação: '+op.compra_venda)

        # Anotamos a nova situação no histórico
        self.historico.append(
            MomentoPosicao(
                op.chave,
                data_pregao,
                self.quantidade,
                self.valor_unitario,
                self.rendimento + rendimento_agregado
            )
        )

        # Retornamos se a posição atual está aberta
        return self.aberta

    def narrativa(self, _print=True) -> Union[None,str]:
        story = [
            f'Posição do título {self.titulo} na corretora {self.corretora}',
            f'\tAberta em {self.data_abertura}' + (
                ', e permanece aberta.'
                if self.aberta else
                f', e fechada em {self.data_fechamento}'
            )
        ]
        for mp in self.historico:
            _, cnpj, nota, _, _, tipo, quantidade, preco = mp.chave_operacao.split(':')
            story += [
                f'\t- {tipo.upper()} {quantidade} X R${printreal(float(preco))}',
                f'\t\tNota de Corretagem: {nota} de {mp.data}',
                f'\t\tSituação: {round(mp.quantidade,2)} X R${printreal(mp.valor_unitario)} = R${printreal(mp.valor_total)} (R${printreal(mp.rendimento)})'
            ]
        story = '\n'.join(story)
        if _print: print(story)
        else: return story

    def sumario(self, _print=True) -> Union[None,str]:
        story = [f'Posição do título {self.titulo} sob custódia da corretora CNPJ {self.corretora},',
                 f'aberta em {self.data_abertura},']
        if self.aberta:
            qtd = self.quantidade
            val = round(self.valor_unitario,2)
            story += [
                f'composta por {int(qtd)} unidades',
                f'de valor médio R${printreal(val)},',
                f'totalizando R${printreal(qtd*val)}.',
            ]
        else:
            story += [
                f'fechada em {self.data_fechamento},',
                f'com rendimento total de R${printreal(self.rendimento)}.'
            ]
        story = ' '.join(story)
        if _print: print(story)
        else: return story



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2017 jfrfonseca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Constantes reutilizadas nos sub-pacotes"""


"""
# CONFIG
"""


# Enderecos das definicioes WSDL
wsdl_cvm = 'http://sistemas.cvm.gov.br/webservices/Sistemas/SCW/CDocs/WsDownloadInfs.asmx?WSDL'
wsdl_bcb = 'https://www3.bcb.gov.br/sgspub/JSP/sgsgeral/FachadaWSSGS.wsdl'

# Codigos do sistema CVM
codigo_informes_diarios_fundos = '209'

# Codigos do sistema BCB (lemos a tabela de forma transposta)
proximidade_nome_indice = 0.95
df_tabela_codigos_bcb = None



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2017 jfrfonseca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Recuperacao de dados de indices a partir do sistema BCB SGS"""


"""
# IMPORTS
"""


# Biblioteca padrão
import difflib

# Dependencias PIP
from suds.client import Client, WebFault  # Recuperacao de dados
import xmltodict  # Parsing dos dados
import numpy as np  # Parsing dos dados
import pandas as pd  # Parsing dos dados
from unidecode import unidecode  # Parsing dos dados

# Este pacote
from py_financas import constantes, _utils


"""
# UTIL
"""


def normaliza_xml(xml):

    # Parseamos o XML como um dicionario pyhton, e identificamos so os valores desejados
    resultado = xmltodict.parse(xml)['SERIES']['SERIE']['ITEM']

    # Normalizamos o tipo dos resultados desejados como lista e retornamos
    if not isinstance(resultado, list):
        return [resultado]
    else:
        return resultado


def _inicializa_tabela_bcb():

    # Inicializamos um dataframe com a tabela de codigos BCB
    df = pd.read_csv('../py_financas/bcb_sgs/tabela_de_codigos_sistemas_web_banco_central_do_brasil.csv',
                     sep=',', header=None, index_col=0).T

    # Montamos diversas variacoes da grafia das strings passadas no arquov CSV
    result = {}
    for coluna in df.columns:

        # Removemos a calitalizacao e os acentos, elementos duplicados e valores nulos, exportando como tipo lista
        result[coluna] = df[coluna].dropna().apply(lambda stri: unidecode(stri.lower())).drop_duplicates().tolist()

    # Salvamos o resultado num DataFrame do modulo de constantes, apos parsear o dicionario acima
    constantes.df_tabela_codigos_bcb = pd.DataFrame().from_dict(result, orient='index').T.fillna(value=np.NaN)


def normaliza_codigo_indice(nome_indice):

    # Inicializamos a tabela de codigos BCB, caso a mesma ainda nao exista
    if constantes.df_tabela_codigos_bcb is None:
        _inicializa_tabela_bcb()

    # Se o nome do indice for um dos codigos da tabela
    if nome_indice in constantes.df_tabela_codigos_bcb.columns:

        # Normalizamos o tipo do codigo indice como inteiro
        codigo_indice = int(nome_indice)

    else:
        # Formatamos o nome passado como uma string sem capitalizacao nem acentos
        nome_indice = unidecode(str(nome_indice).lower())
        try:
            # Identificamos a primeira coluna do dataframe que contenha o nome exato informado
            codigo_indice = int(next(coluna for coluna in constantes.df_tabela_codigos_bcb.columns
                                     if nome_indice in constantes.df_tabela_codigos_bcb[coluna]))
        except StopIteration:
            # No caso de termos um erro de fim de iteração, nenhuma das colunas do dataframe tem um match exato :(
            # Portanto, sera necessario fazer um match aproximado
            try:
                codigo_indice = int(next(
                    coluna for coluna in constantes.df_tabela_codigos_bcb.columns

                    # Recuperamos a primeira coluna que tenha algum elemento parecido o bastante com o nome requisitado
                    if len(difflib.get_close_matches(nome_indice,
                                                     constantes.df_tabela_codigos_bcb[coluna].dropna(),
                                                     n=1, cutoff=constantes.proximidade_nome_indice)) > 0
                ))
            except StopIteration:

                # Nao conseguimos encontrar o codigo do indice e nao temos mais alternativas :( Interrompemos a funcao
                raise KeyError("Nao foi possivel identificar o indice {}".format(nome_indice))
    return codigo_indice


"""
# PAYLOAD
"""


def recupera_indice(nome_indice, data_inicio, data_fim=_utils.ultimo_dia_util()):

    # Validamos a entrada
    assert data_inicio <= data_fim, "Data {} nao e anterior a data {}".format(data_inicio, data_fim)

    # Normalizamos o nome do indice recebido como uma lista
    if not isinstance(nome_indice, list):
        nome_indice = [nome_indice]

    # Traduzimos os codigos dos bcb_sgs buscados
    lista_codigos_indices = [normaliza_codigo_indice(indice) for indice in nome_indice]

    # Inicializamos um cliente SUDS WSDL apontado para o arquivo WSDL da plataforma BCB SGS
    server = Client(constantes.wsdl_bcb)

    # Inicializamos um buffer para os resultados
    buffer_series = []

    # Para cada codigo de indice a ser recuperado
    for codigo_indice, nome_recebido_indice in zip(lista_codigos_indices, nome_indice):
        try:
            string_xml = server.service.getValoresSeriesXML([int(codigo_indice)],
                                                            data_inicio.strftime('%d/%m/%Y'),
                                                            data_fim.strftime('%d/%m/%Y'))

            # Formatamos o resultado do codigo atual
            valores = normaliza_xml(string_xml)

            # Formatamos a lista de valores recuperada como uma Serie Pandas
            serie = pd.Series(pd.to_numeric([v['VALOR'] for v in valores]),
                              index=pd.to_datetime([v['DATA'] for v in valores], dayfirst=True),
                              name=nome_recebido_indice)
            buffer_series.append(serie)

        # No caso de nao termos recuperado nenhum dado para o indice atual, o ignoramos
        except WebFault as wbf_i:
            if 'Value(s) not found' in str(wbf_i):
                pass
            else:
                raise wbf_i

    # Apos o loop, unificamos as series como um dataframe que retornamos
    return pd.DataFrame(buffer_series).T





# STDLib
import re
from datetime import datetime
from typing import Dict, List, Union, Tuple, Set

# PIP
import pdfplumber
import pandas as pd
from numpy import nan

# This package
from py_financas.sinacor.types import to_DC, to_float, \
    NotaCorretagem, Operacao, Corretora, Impostos, Custos, Totais


# CONSTANTS
#######################################################################################################################


date = '[0-9]{2}/[0-9]{2}/20[0-9]{2}'
cnpj = '[0-9]{2}\.[0-9]{3}\.[0-9]{3}/0001-[0-9]{2}'
text = '[a-zçáãâéêíóôõúA-ZÇÁÃÂÉÊÍÓÕÔÚ /,\.]'
ignore_line = '[^\n]+\n'
num = '[0-9]{1,3}(?:\.[0-9]{3})*(?:,[0-9]{2})?'
Q = '(?:([^\s]+)\s+)?'
NEGOCIACAO = '((?:1-)?BOVESPA(?: 1)?)'
CV = '\s+(C|V)'
TIPO_MERCADO = '\s+(VISTA|FRACIONARIO)'
PRAZO = '(?:\s+([^\s]+))?'
TITULO = '\s+(.*)'
TIPO_TITULO = '\s+(CI|ON|PN)(?:[\s\t]*(E[D]?[JR]?))?(?:[\s\t]*(ER|ES|ON|NM|N[0-9]))?'
OBS = '(?:\s+([^\s]+))?'
QUANTIDADE = f'\s+({num})'
PRECO_AJUSTE = f'\s+({num})'
VALOR_OPERACAO_AJUSTE = f'\s+({num})'
DC = '\s+(D|C)'

rgx_header = re.compile(
    '^NOTA DE CORRETAGEM\n'
    +ignore_line
    +f'([0-9]+) ([0-9]+) ({date})\n'  # Numero Nota, Numero Folha, Data Pregao
    +f'({text}+)\n'  # Nome Corretora
    +ignore_line
    +'Tel'+ignore_line
    +'Internet'+ignore_line
    +f'[CNPJ\.:\s\t]+({cnpj})[^\n]+\n'  # CNPJ Corretora
    +'Ouvidoria'+ignore_line
    +'Cliente'+ignore_line
    +'([^\n]+)\n'  # Cliente
)

rgx_body = re.compile(
    'Negócios realizados\n'
    +'Q'+ignore_line
    +'((.*\n)+)'
    +'Resumo dos'+ignore_line
)

rgx_op = re.compile(
    f'^{Q}{NEGOCIACAO}{CV}{TIPO_MERCADO}{TITULO}{TIPO_TITULO}{OBS}{QUANTIDADE}{PRECO_AJUSTE}{VALOR_OPERACAO_AJUSTE}{DC}$'
)

rgx_footer = re.compile(
    f'Resumo dos Neg.*\n'
    f'Debêntures ({num}) Clearing\n'
    f'Vendas à vista ({num}) Valor líquido das operações ({num}){DC}\n'
    f'Compras à vista ({num}) Taxa de liquidação ({num}){DC}\n'
    f'Opções - compras ({num}) Taxa de Registro ({num}){DC}\n'
    f'Opções - vendas ({num}) Total CBLC ({num}){DC}\n'
    f'Operações à termo ({num}) Bolsa\n'
    f'Valor [^\n]+ títulos [^\n]+ ({num}) Taxa de [^\n]+ ({num}){DC}\n'
    f'Valor das operações ({num}) Taxa [^\n]+ ({num}){DC}\n'
    f'Emolumentos ({num}){DC}\n'
    f'Total Bovespa [^\n]+ ({num}){DC}\n'
    f'Especific[^\n]+\n'
    f'Clearing ({num}){DC}\n'
    f'A coluna Q indica [^\n]+ ({num}){DC}\n'
    f'(?:Esta nota substitui a nota nº ([0-9]+) )?Execução casa ({num}){DC}\n'
    f'ISS \(\s*(RIO DE JANEIRO|SÃO PAULO)\s*\) ({num})(?:{DC})?\n'
    f'(?:I\.R\.R\.F\. s. operações, base .. ({num}) ({num})(?:{DC})?\n)?'
    f'Outras ({num}){DC}\n'
    f'Total corretagem / Despesas ({num}){DC}\n'
    f'\(\*\) - Observações: A - Posição Futuro T - Liquidação pelo Bruto Líquido para ({date}) ({num}){DC}\n'
)


# UTILS
#######################################################################################################################


def read_pages(all_files:List[str]) -> List[str]:
    all_pages = []
    for filename in all_files:
        with pdfplumber.open(filename) as pdf:
            for page in pdf.pages:
                all_pages.append(page.extract_text())
    return all_pages


def validate_matches(valid_pages, matches):
  validation = [((not vp) or (m is not None))  # VP -> M (Modus Ponens)
                for vp,m in zip(valid_pages, matches)]
  if not all(validation):
      invalid = [i for i,val in enumerate(validation) if not val]
      raise AssertionError(f'SOME VALID PAGES WERE NOT PARSED! {invalid}')


def check_continuation(page):
    return 'C O N T I N U A . . .' in page


# PARSERS
#######################################################################################################################


def parse_header(all_pages:List[str], valid_pages:List[bool]) -> List[Dict[str,str]]:

    # Extract and validate the header sections in the pages
    header_sections = [
        m
        if ((m := rgx_header.findall(p)) is None)
        else (m[0] if (len(m) > 0) else None)
        for p in all_pages
    ]
    validate_matches(valid_pages, header_sections)

    # Parse the header objects
    headers = [
        {} if i is None else dict(zip([
            'nota_numero', 'nota_folha', 'nota_data',
            'corretora_nome', 'corretora_cnpj',
            'cliente'
        ], list(i)))
        for i in header_sections
    ]
    return headers


def parse_body(all_pages:List[str], valid_pages:List[bool]) -> List[List[Union[Dict[str,str], None]]]:

    # Extract and validate the body sections in the pages
    body_sections = [
        m
        if ((m := rgx_body.findall(p)) is None)
        else (m[0] if (len(m) > 0) else None)
        for p in all_pages
    ]
    body_sections = [b if (b is None) else b[0].splitlines() for b in body_sections]
    validate_matches(valid_pages, body_sections)

    # Extract and validate the operations sections in the body sections
    operation_sections = [
        i if (i is None) else [
            rgx_op.findall(j)
            for j in i
        ]
        for i in body_sections
    ]
    validate_matches(valid_pages, operation_sections)

    # Parse the body objects
    body_columns = [
        'q', 'negociacao', 'cv_compra_venda', 'tipo_mercado', # 'prazo',
        'especificacao_titulo_0',
        'especificacao_titulo_1',
        'especificacao_titulo_2',
        'especificacao_titulo_3',
        'observacao',
        'val_quantidade', 'val_preco_ajuste', 'val_valor_operacao_ajuste',
        'dc_debito_credito',
    ]
    body = [
        None if b is None else [dict(zip(body_columns, list(*l))) for l in b]
        for b in operation_sections
    ]
    validate_matches(valid_pages, body)

    # Validate the parsed operations in the body objects
    for page,b in enumerate(body):
        if b is None: continue
        for num_op,op in enumerate(b):
            assert len(op) == len(body_columns), f'Could not parse operation {num_op} in page {page} [\n\t{len(op)}\n\t{op}\n]'

    return body  #type:ignore


def parse_footer(all_pages:List[str], valid_pages:List[bool]) -> Tuple[List[Union[Dict[str,str],None]],List[bool]]:

    # Extract and validate the footer sections in the pages
    footer_sections = [
        m
        if ((m := rgx_footer.findall(p)) is None)
        else (m[0] if (len(m) > 0) else None)
        for p in all_pages
    ]
    footer_continuation = [check_continuation(p) for p in all_pages]
    footer_sections = [('' if cont else m) if (m is None) else m
                       for m,cont in zip(footer_sections, footer_continuation)]
    validate_matches(valid_pages, footer_sections)

    # Parse the footer objects
    footer = [
        None if i in {'', None} else dict(zip([
            'val_debentures',
            'val_vendas_vista', 'val_valor_liquido_operacoes', 'dc_valor_liquido_operacoes',
            'val_compras_vista', 'val_taxa_liquidacao', 'dc_taxa_liquidacao',
            'val_opcoes_compras', 'val_taxa_registro', 'dc_taxa_registro',
            'val_opcoes_vendas', 'val_total_cblc', 'dc_total_cblc',
            'val_operacoes_termo_bolsa',
            'val_operacoes_titulos_publicos', 'val_taxa_termo_opcoes', 'dc_taxa_termo_opcoes',
            'val_operacoes', 'val_taxa_ana', 'dc_taxa_ana',
            'val_emolumentos', 'dc_emolumentos',
            'val_total_bovespa_soma', 'dc_total_bovespa_soma',
            'val_clearing', 'dc_clearing',
            'val_execucao', 'dc_execucao',
            'nota_substituida', 'val_execucao_casa', 'dc_execucao_casa',
            'corretora_estado', 'val_iss', 'dc_iss',
            'val_base_irrf', 'val_irrf', 'dc_irrf',
            'val_outras', 'dc_outras',
            'val_total_corretagem_despesas', 'dc_total_corretagem_despesas',
            'nota_data_liquido', 'val_liquido', 'dc_liquido',
        ], list(i)))  #type:ignore
        for i in footer_sections
    ]
    # Do not use the same validation because some pages have continuation pages
    #validate_matches(valid_pages, footer)

    return footer, footer_continuation


def _base_parse_notas_corretagem(all_pages:List[str]) -> Tuple[List[Union[Dict[str,str],None]],
                                                               Set[str],Set[str],Set[str]]:

    # Prepare the buffers
    valid_pages = [(len(p.strip()) > 300) for p in all_pages]
    notas_corretagem = []
    header_columns = None
    body_columns = None
    footer_columns = None

    # Parse the sections
    header = parse_header(all_pages, valid_pages)
    body = parse_body(all_pages, valid_pages)
    footer, footer_continuation = parse_footer(all_pages, valid_pages)

    # Join as a single object
    pos = -1
    for h,b,f,cont,p in zip(header,body,footer,footer_continuation,all_pages):
        pos += 1

        # Skip invalid pages
        if len(h) == 0:
            assert b is None, 'OPERATIONS ON AN EMPTY HEADER'
            assert f is None, 'FOOTER ON AN EMPTY HEADER'
            notas_corretagem.append(None)
            continue

        # Assert column definition stability
        if header_columns is None:
            header_columns = set(h.keys())
        else: assert header_columns == set(h.keys())
        if body_columns is None:
            body_columns = set(b[0].keys())  #type:ignore
        else: assert body_columns == set(b[0].keys())  #type:ignore
        if not cont:
            if footer_columns is None:
                footer_columns = set(f.keys())  #type:ignore
            else: assert footer_columns == set(f.keys())  #type:ignore

        # Join the data
        if not cont:
            h.update(f)  #type:ignore
        assert not (cont and (f is not None)), 'IS CONTINUATION BUT HAS FOOTER'
        assert not ((not cont) and (f is None)), 'IS NOT CONTINUATION BUT HAS NO FOOTER'
        h.update({'dataset_page': pos, 'raw_page': p, 'has_continuation': cont, 'operations': b})  #type:ignore
        notas_corretagem.append(h)

    # Validate the entire dataset
    validate_matches(valid_pages, notas_corretagem)
    return notas_corretagem, header_columns, body_columns, footer_columns  #type:ignore


def parse_notas_corretagem(all_files:List[str]) -> List[NotaCorretagem]:

    # Read the files as strings
    all_pages = read_pages(all_files)

    # Base-parse and validate the file contents
    notas_corretagem, header_columns, body_columns, footer_columns = _base_parse_notas_corretagem(all_pages)

    # Format the individually-validated Notas de Corretagem as a dataframe
    df_notas_corretagem = pd.DataFrame([n for n in notas_corretagem if n is not None])

    # Group by Número da Nota and validate the many pages of the same Nota de Corretagem
    obj_notas_corretagem = []
    for nota_numero, sdf in df_notas_corretagem.groupby('nota_numero'):
        if len(sdf) == 1:
            lp_df = sdf

        # Validate this multi-page Nota de Corretagem\
        else:

            # Each page should appear just once
            assert len(sdf) == len(sdf['nota_folha'].unique()), f'REPEATED PAGES IN NOTA {nota_numero}'

            # All the values in each Header column must be equal
            for col in header_columns:
                if col == 'nota_folha': continue  # No need to validate the number of the page
                values = sdf[col].unique()
                assert len(values) == 1, f'DIFFERENT VALUES FOR HEADER COLUMN {col} OF NOTA {nota_numero}: {values}'

            # All pages must have operations
            for _, row in sdf.iterrows():
                nota_folha = row['nota_folha']
                assert len(row['operations']) > 0, f'NO OPERATIONS ON PAGE {nota_folha} OF NOTA {nota_numero}'

            # Only a single page (the last) should gave the footer
            last_page = sdf['nota_folha'].astype(int).max()
            assert last_page > 1, f'SINGLE PAGE IN NOTA {nota_numero}, THAT HAS {len(sdf)} ROWS!'
            null_df = sdf[sdf['nota_folha'] != str(last_page)][list(footer_columns)]
            lp_df = sdf[sdf['nota_folha'] == str(last_page)][list(footer_columns)]
            assert len(lp_df) == 1, f'MORE THAN ONE LAST PAGE {last_page} IN NOTA {nota_numero}'
            for col in footer_columns:
                null_values = set(null_df[col].unique())
                assert null_values == {nan}, f'NON-NULL VALUES IN FOOTER COLUMN {col} OF THE MIDDLE PAGES OF NOTA {nota_numero}: {null_values}'
                assert not lp_df[col].isnull().all(), f'NULL VALUE IN FOOTER COLUMN {col} OF THE LAST PAGE OF NOTA {nota_numero}'

        # This Nota de Corretagem has passed the group validations. Convert it to a NotaCorretagem object
        lr = lp_df.iloc[0]
        dataset_page = str(sdf.iloc[0]['dataset_page'])
        try:

            # Header
            _raw = sdf.iloc[0]['raw_page'].strip()
            nota_numero = int(sdf.iloc[0]['nota_numero'].strip())
            numero_nota_substitutiva = lp_df['nota_substituida'].fillna('').iloc[0].strip()
            if numero_nota_substitutiva == '': numero_nota_substitutiva = -1
            numero_nota_substitutiva = int(numero_nota_substitutiva)
            nota_data_pregao = datetime.strptime(sdf.iloc[0]['nota_data'].strip(), '%d/%m/%Y').date()
            nota_data_liquidacao = datetime.strptime(lr['nota_data_liquido'].strip(), '%d/%m/%Y').date()
            cliente = sdf.iloc[0]['cliente'].strip()
            cnpj = sdf.iloc[0]['corretora_cnpj'].strip()
            estado = lp_df['corretora_estado'].iloc[0].lower().strip()

            # Corretora
            corretora = Corretora(
                nome=sdf.iloc[0]['corretora_nome'].strip(),
                cnpj=str(''.join(i for i in cnpj if i.isdigit())).split('.')[0],
                estado={'são paulo': 'SP', 'rio de janeiro': 'RJ'}[estado]  #type:ignore
            )

            # Footer
            taxes = Impostos(
                iss=to_float(lr['val_iss']),
                iss_DC=to_DC(lr['dc_iss'], 'debito'),  #type:ignore
                irrf=to_float(lr['val_irrf']),
                irrf_base=to_float(lr['val_base_irrf']),
                irrf_DC=to_DC(lr['dc_irrf'], 'debito')  #type:ignore
            )
            costs = Custos(
                taxa_liquidacao=to_float(lr['val_taxa_liquidacao']),
                taxa_liquidacao_DC=to_DC(lr['dc_taxa_liquidacao'], 'debito'),  #type:ignore
                taxa_registro=to_float(lr['val_taxa_registro']),
                taxa_registro_DC=to_DC(lr['dc_taxa_registro'], 'debito'),  #type:ignore
                taxa_termo_opcoes=to_float(lr['val_taxa_termo_opcoes']),
                taxa_termo_opcoes_DC=to_DC(lr['dc_taxa_termo_opcoes'], 'debito'),  #type:ignore
                taxa_ana=to_float(lr['val_taxa_ana']),
                taxa_ana_DC=to_DC(lr['dc_taxa_ana'], 'debito'),  #type:ignore
                emolumentos=to_float(lr['val_emolumentos']),
                emolumentos_DC=to_DC(lr['dc_emolumentos'], 'debito'),  #type:ignore
                clearing=to_float(lr['val_clearing']),
                clearing_DC=to_DC(lr['dc_clearing'], 'debito'),  #type:ignore
                execucao=to_float(lr['val_execucao']),
                execucao_DC=to_DC(lr['dc_execucao'], 'debito'),  #type:ignore
                execucao_casa=to_float(lr['val_execucao_casa']),
                execucao_casa_DC=to_DC(lr['dc_execucao_casa'], 'debito'),  #type:ignore
                outras=to_float(lr['val_outras']),
                outras_DC=to_DC(lr['dc_outras'], 'debito'),  #type:ignore
            )
            totals = Totais(
                liquido=to_float(lr['val_liquido']),
                liquido_DC=to_DC(lr['dc_liquido'], 'credito'),  #type:ignore
                vendas_vista=to_float(lr['val_vendas_vista']),
                compras_vista=to_float(lr['val_compras_vista']),
                opcoes_compras=to_float(lr['val_opcoes_compras']),
                opcoes_vendas=to_float(lr['val_opcoes_vendas']),
                debentures=to_float(lr['val_debentures']),
                operacoes=to_float(lr['val_operacoes']),
                operacoes_termo_bolsa=to_float(lr['val_operacoes_termo_bolsa']),
                operacoes_titulos_publicos=to_float(lr['val_operacoes_titulos_publicos']),
                valor_liquido_operacoes=to_float(lr['val_valor_liquido_operacoes']),
                valor_liquido_operacoes_DC=to_DC(lr['dc_valor_liquido_operacoes'], 'credito'),  #type:ignore
                cblc=to_float(lr['val_total_cblc']),
                cblc_DC=to_DC(lr['dc_total_cblc'], 'credito'),  #type:ignore
                bovespa_soma=to_float(lr['val_total_bovespa_soma']),
                bovespa_soma_DC=to_DC(lr['dc_total_bovespa_soma'], 'credito'),  #type:ignore
                corretagem_despesas=to_float(lr['val_total_corretagem_despesas']),
                corretagem_despesas_DC=to_DC(lr['dc_total_corretagem_despesas'], 'credito'),  #type:ignore
            )

            # Operations
            operations = []
            for _, row in sdf.iterrows():
                operations.append([])
                for op in row['operations']:
                    operations[-1].append(Operacao(
                        _nota_numero=nota_numero, _nota_data_pregao=nota_data_pregao,
                        _nota_corretora_cnpj=str(corretora.cnpj).split('.')[0],
                        titulo=op['especificacao_titulo_0'].strip(),
                        preco_ajuste=to_float(op['val_preco_ajuste']),
                        quantidade=to_float(op['val_quantidade']),
                        valor_operacao_ajuste=to_float(op['val_valor_operacao_ajuste']),
                        compra_venda={'C': 'compra', 'V': 'venda'}[op['cv_compra_venda'].strip()],  #type:ignore
                        debito_credito=to_DC(op['dc_debito_credito']),  #type:ignore
                        tipo_mercado=op['tipo_mercado'].strip().lower(),
                        negociacao=op['negociacao'].strip().lower(),
                        especificacao_titulo=(
                            op['especificacao_titulo_1'].strip().upper(),
                            op['especificacao_titulo_2'].strip().upper(),
                            op['especificacao_titulo_3'].strip().upper()),
                        observacao=op['observacao'].strip()
                    ))

            # Nota de Corretagem
            obj_notas_corretagem.append(
                NotaCorretagem(
                    _raw_content=_raw, numero=nota_numero,
                    data_pregao=nota_data_pregao, data_liquidacao=nota_data_liquidacao,
                    numero_nota_substitutiva=numero_nota_substitutiva,
                    corretora=corretora, cliente=cliente, quantidade_folhas=len(sdf),
                    operacoes=operations, impostos=taxes, custos=costs, totais=totals
                )
            )

        except Exception as exp:
            raise Exception(f'On page {dataset_page}:\n\t{exp}')

    return obj_notas_corretagem



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2017 jfrfonseca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Operacoes com arquivos ZIP"""


"""
# IMPORTS
"""


# Biblioteca padrão
import zipfile
import tempfile


"""
# UTILS
"""


def le_arquivo_zip_de_string(string_zipada):
    arquivo_temporario = tempfile.TemporaryFile()
    arquivo_temporario.write(string_zipada)
    string_zipada_zip = zipfile.ZipFile(arquivo_temporario, "r")
    return string_zipada_zip.read(string_zipada_zip.filelist[0].filename)
